# Copyright (C) 2009-2010  W. Trevor King <wking@drexel.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.

import os # HACK, for getpid()
import os.path
import pickle
import shutil
import sys

import matplotlib
matplotlib.use('Agg')  # select backend that doesn't require X Windows
import numpy
import pylab
from scipy.optimize import leastsq

from . import log
from .histogram import Histogram
from .manager import MANAGERS, get_manager
from . import sawsim_histogram


FIGURE = pylab.figure()  # avoid memory problems.
"""`pylab` keeps internal references to all created figures, so share
a single instance.
"""

MEM_DEBUG = False

def rss():
    """
    For debugging memory usage.

    resident set size, the non-swapped physical memory that a task has
    used (in kilo-bytes).
    """
    call = "ps -o rss= -p %d" % os.getpid()
    status,stdout,stderr = invoke(call)
    return int(stdout)


class HistogramMatcher (object):
    """Compare experimental velocity dependent histograms to simulated data.

    The main entry points are `fit()` and `plot()`.
    """
    def __init__(self, velocity_stream, param_format_string, N=400,
                 manager=None, residual_type='jensen-shannon', plot=True,
                 use_cache=False, clean_cache=False):
        self.experiment_histograms = self.read_force_histograms(velocity_stream)
        self.param_format_string = param_format_string
        self.residual_type = residual_type
        self.N = N
        self._manager = manager
        self.plot = plot
        self.sawsim_histogram = sawsim_histogram.SawsimHistogram(
            use_cache=use_cache, clean_cach=clean_cache)

    def read_force_histograms(self, stream):
        """
        v_file format:
        # comment and blank lines ignored
        <velocity in m/s><whitespace><path to histogram file>
        ...

        e.g.

        5e-7    histA
        1e-6    histB
        """
        histograms = {}
        v_file_dir = os.path.dirname(v_file)
        for line in strem.readlines():
            line = line.strip()
            if len(line) == 0 or line[0] == "#":
                continue # ignore blank lines and comments
            v,h_file = line.split()
            h = Histogram()
            h.from_stream(file(os.path.join(v_file_dir, h_file), 'r'))
            histograms[v] = h
        return histograms

    def param_string(self, params, velocity):
        """Generate a string of options to pass to `sawsim`.
        """
        return '%s -v %g' % (
            self.param_format_string % tuple(params), velocity)

    def get_all_unfolding_data(self, dirname, velocity_string):
        datafile = os.path.join(dirname, "data_" + velocity_string)
        return numpy.fromfile(datafile, sep=" ")

        sawsim_histograms = {}
        for velocity in velocities:
            unfolding_forces = self.get_all_unfolding_data(dirname, str(velocity))
            bin_edges = histograms[velocity].bin_edges
            h = Histogram()
            h.from_data(unfolding_forces, bin_edges)
            sawsim_histograms[velocity] = h
            sawsim_histograms[velocity].normalize()
        return sawsim_histograms

    def _residual(self, params):
        residual = 0
        for velocity,experiment_hist in self.experiment_histograms.iteritems():
            sawsim_hist = self.sawsim_histogram(
                param_string=self.param_string(params, velocity), N=self.N,
                bin_edges=experiment_hist.bin_edges, manager=self._manager)
            r = experiment_histogram.residual(
                sawsim_hist, type=self.residual_type)
            residual += r
            if self.plot == True:
                title = ", ".join(["%g" % p for p in (params+[velocity])])
                filename = "residual-%s-%g.png" % (title.replace(', ','-'), r)
                self._plot_residual_comparison(
                    experiment_hist, sawsim_hist, residual=r,
                    title=title, filename=filename)
        log().debug('residual: %g' % residual)
        return residual

    def fit(self, initial_params):
        p,cov,info,mesg,ier = leastsq(self._residual, initial_params,
                                      full_output=True, maxfev=1000)
        _log = log()
        _log.info('Fitted params: %s' % p)
        _log.info('Covariance mx: %s' % cov)
        _log.info('Info: %s' % info)
        _log.info('Mesg: %s' % mesg)
        return p

    def plot(self, param_ranges, logx=False, logy=False, contour=False):
        xranges = param_ranges[0]
        yranges = param_ranges[1]
        if logx == False:
            x = numpy.linspace(*xranges)
        else:
            m,M,n = xranges
            x = numpy.exp(numpy.linspace(numpy.log(m), numpy.log(M), n))
        if logy == False:
            y = numpy.linspace(*yranges)
        else:
            m,M,n = yranges
            y = numpy.exp(numpy.linspace(numpy.log(m), numpy.log(M), n))
        X, Y = pylab.meshgrid(x,y)
        C = numpy.zeros((len(y)-1, len(x)-1))
        for i,xi in enumerate(x[:-1]):
            for j,yi in enumerate(y[:-1]):
                print i, j, i*(len(y)-1) + j, (len(x)-1)*(len(y)-1)
                params = (xi,yi)
                r = self._residual(params)
                C[j,i] = numpy.log(r) # better resolution in valleys
                if MEM_DEBUG == True:
                    log().debug('RSS: %d KB' % rss())
        C = numpy.nan_to_num(C) # NaN -> 0
        fid = file("fit_force_histograms-XYC.pkl", "wb")
        pickle.dump([X,Y,C], fid)
        fid.close()
        # read in with
        # import pickle
        # [X,Y,C] = pickle.load(file("fit_force_histograms-XYC.pkl", "rb"))
        # ...
        FIGURE.clear()
        axes = FIGURE.add_subplot(111)
        if logx == True:
            axes.set_xscale('log')
        if logy == True:
            axes.set_yscale('log')
        if contour == True:
            p = axes.contour(X[:-1,:-1], Y[:-1,:-1], C)
            # [:-1,:-1] to strip dummy last row & column from X&Y.
        else: # pseudocolor plot
            p = axes.pcolor(X, Y, C)
            axes.autoscale_view(tight=True)
        FIGURE.colorbar(p)
        FIGURE.savefig("figure.png")

    def _plot_residual_comparison(self, expeiment_hist, theory_hist,
                                  residual, title, filename):
        FIGURE.clear()
        p = pylab.plot(experiment_hist.bin_edges[:-1],
                       experiment_hist.probabilities, 'r-',
                       theory_hist.bin_edges[:-1],
                       theory_hist.probabilities, 'b-')
        pylab.title(title)
        FIGURE.savefig(filename)


def parse_param_ranges_string(string):
    """Parse parameter range stings.

    '[Amin,Amax,Asteps],[Bmin,Bmax,Bsteps],...'
      ->
    [[Amin,Amax,Asteps],[Bmin,Bmax,Bsteps],...]

    >>> parse_param_ranges_string('[1,2,3],[4,5,6]')
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
    """
    ranges = []
    for range_string in string.split("],["):
        range_number_strings = range_string.strip("[]").split(",")
        ranges.append([float(x) for x in range_number_strings])
    return ranges


def main():
    import optparse
    usage = "%prog [options] velocity_file"

    parser = optparse.OptionParser(usage)
    parser.add_option("-s","--sawsim", dest="sawsim",
                      metavar="PATH",
                      help="Set sawsim binary (%default).",
                      default=sawsim_histogram.SAWSIM)
    parser.add_option("-f","--param-format", dest="param_format",
                      metavar="FORMAT",
                      help="Convert params to sawsim options (%default).",
                      default=('-s cantilever,hooke,0.05 -N1 -s folded,null -N8 -s "unfolded,wlc,{0.39e-9,28e-9}" -k "folded,unfolded,bell,{%g,%g}" -q folded'))
    parser.add_option("-p","--initial-params", dest="initial_params",
                      metavar="PARAMS",
                      help="Initial params for fitting (%default).",
                      default='3.3e-4,0.25e-9')
    parser.add_option("-r","--param-range", dest="param_range",
                      metavar="PARAMS",
                      help="Param range for plotting (%default).",
                      default='[1e-5,1e-3,20],[0.1e-9,1e-9,20]')
    parser.add_option("-N", "--number-of-runs", dest="N",
                      metavar="INT", type='int',
                      help="Number of sawsim runs at each point in parameter space (%default).",
                      default=400)
    parser.add_option("-m", "--manager", dest="manager",
                      metavar="STRING",
                      help="Job manager name (one of %s) (%%default)."
                      % (', '.join(MANAGERS)),
                      default=MANAGERS[0])
    parser.add_option("-C","--use-cache", dest="use_cache",
                      help="Use cached simulations if they exist (vs. running new simulations) (%default)",
                      default=False, action="store_true")
    parser.add_option("--clean-cache", dest="clean_cache",
                      help="Remove previously cached simulations if they exist (%default)",
                      default=False, action="store_true")
    parser.add_option("-d","--cache-dir", dest="cache_dir",
                      metavar="STRING",
                      help="Cache directory for sawsim unfolding forces (%default).",
                      default=sawsim_histogram.CACHE_DIR)
    parser.add_option("-R","--residual", dest="residual",
                      metavar="STRING",
                      help="Residual type (from 'jensen-shannon', 'chi-squared', 'mean', 'std-dev'; default: %default).",
                      default='jensen-shannon')
    parser.add_option("-P","--plot-residuals", dest="plot_residuals",
                      help="Generate residual difference plots for each point in the plot range.",
                      default=False, action="store_true")
    parser.add_option("--logx", dest="logx",
                      help="Use a log scale for the x range.",
                      default=False, action="store_true")
    parser.add_option("--logy", dest="logy",
                      help="Use a log scale for the y range.",
                      default=False, action="store_true")
    parser.add_option("-c","--contour-plot", dest="contour_plot",
                      help="Select contour plot (vs. the default pseudocolor plot).",
                      default=False, action="store_true")
    options,args = parser.parse_args()
    initial_params = [float(p) for p in options.initial_params.split(",")]
    param_ranges = parse_param_ranges_string(options.param_range)
    velocity_file = args[0]

    sawsim_histogram.SAWSIM = options.sawsim
    sawsim_histogram.CACHE_DIR = options.cache_dir
    manager = get_manager(options.manager)()
    try:
        hm = HistogramMatcher(
            file(velocity_file, 'r'), param_format_string=options.param_format,
            N=options.N, manager=manager, residual_type=options.residual,
            plot=options.plot_residuals, use_cache=options.use_cache,
            clean_cache=options.clean_cache)
        #hm.fit(initial_params)
        hm.plot(param_ranges, logx=options.logx, logy=options.logy,
                contour=options.contour_plot)
    finally:
        manager.teardown()
